#ifdef KEVIL

/* The format of kevil log:
 *
 * entry = <time> [newwin | winid] <key>
 *
 * where
 *
 * time   = 7bit encoded time difference in msec.
 * newwin = newwin_prefix (16383) name_str '\0'
 * winid  = winid_prefix (16382) winid
 * key    = 7bit encoded (keycode | (mod << 8))
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include "evilwm.h"
#include "client.h"

#define WIN_NAME_NUM (256 - 1)

static FILE* fp;
static int disabled;
static long long prev_msec;
static Window prev_win;
static Time key_time[256];
static const char* win_names[WIN_NAME_NUM];
static const int NEWWIN_PREFIX = 16383;
static const int WINID_PREFIX = 16382;

static int kevil_init() {
    int err;
    const char* home = getenv("HOME");
    char buf[256];
    struct timeval tv;
    struct tm* timep;

    if (fp) {
        return 0;
    }

    gettimeofday(&tv, NULL);
    timep = localtime(&tv.tv_sec);

    if (!home) {
        fprintf(stderr, "failed to get home directory\n");
        return -1;
    }
    strcpy(buf, home);
    strcat(buf, "/.sevilwm");
    err = mkdir(buf, S_IRWXU);
    if (err < 0 && errno != EEXIST) {
        perror("failed to create ~/.sevilwm");
        return -1;
    }

    sprintf(buf + strlen(buf), "/%d", timep->tm_year + 1900);
    err = mkdir(buf, S_IRWXU);
    if (err < 0 && errno != EEXIST) {
        perror("failed to create ~/.sevilwm/<year>");
        return -1;
    }

    sprintf(buf + strlen(buf), "/%d%d-%d%d%d",
            timep->tm_mon, timep->tm_mday,
            timep->tm_hour, timep->tm_min, timep->tm_sec);
    fp = fopen(buf, "wb");
    if (!fp) {
        perror("failed to create key log file");
        return -1;
    }

    return 0;
}

static long long get_msec() {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (long long)tv.tv_sec * 1000 + tv.tv_usec / 1000;
}

static void enc_ll(long long ll) {
    while (ll >= 128) {
        int v = ll & 127;
        fputc(v | 128, fp);
        ll >>= 7;
    }
    fputc(ll, fp);
}

static void enc_time() {
    long long msec;
    msec = get_msec();
    enc_ll(msec - prev_msec);
    prev_msec = msec;
}

static void enc_newwin(int id, const char* name) {
    enc_ll(NEWWIN_PREFIX);
    fputs(name, fp);
    fputc(0, fp);
}

static void enc_win(Window w) {
    Client* cl = find_client(w);
    const char* name = cl ? cl->name : "root";
    int i;
    for (i = 0; i < WIN_NAME_NUM; i++) {
        if (win_names[i]) {
            if (!strcmp(win_names[i], name)) {
                break;
            }
        }
        else {
            win_names[i] = name;
            enc_newwin(i, name);
            return;
        }
    }
    if (i == WIN_NAME_NUM) {
        fprintf(stderr, "Too many windows: %s\n", name);
    }

    enc_ll(WINID_PREFIX);
    fputc(i, fp);
}

static void enc_key(XKeyEvent *e) {
    int v = e->keycode | (e->state << 8);
    if (key_time[e->keycode] == e->time) {
        v |= 1 << 16;
    }
    enc_ll(v);
}

void kevil_log(XKeyEvent *e) {
    if (disabled) {
        return;
    }

    if (kevil_init()) {
        fprintf(stderr, "failed to initialize kevil\n");
        disabled = 1;
        return;
    }

    enc_time();

    if (e->window != prev_win) {
        enc_win(e->window);
    }

    enc_key(e);

    fflush(fp);
/*
    printf("%d %d %d\n",
           XKeycodeToKeysym(dpy, e->keycode, 0),
           e->keycode, e->state);
*/
}

void kevil_log_release(XKeyEvent* e) {
    key_time[e->keycode] = e->time;
}

#endif
