#!/usr/bin/env ruby

log = ARGV[0]
if !log
  puts "Usage: #$0 log"
  exit 1
end

def dec_ll(f)
  k = f.getc
  v = s = 0
  while k >= 128
    v |= (k & 127) << s
    s += 7
    k = f.getc
  end
  v |= k << s
end

NEWWIN_PREFIX = 16383
WINID_PREFIX = 16382
MODS = ['Shift', 'Lock', 'Ctrl', 'Mod1', 'Mod2', 'Mod3', 'Mod4', 'Mod5', 'Rep']
KEYSYMS = {}
File.readlines(File.dirname(__FILE__) + '/keysym.txt').each do |l|
  l.chomp!
  k, v = l.split
  KEYSYMS[k.to_i] = v
end

msec = 0
win = nil
win_names = []
File.open(log) do |f|
  while !f.eof?
    msec += dec_ll(f)
    key = dec_ll(f)
    if key == NEWWIN_PREFIX
      n = ''
      k = f.getc
      while k != 0
        n << k
        k = f.getc
      end
      win_names << win = n
      key = dec_ll(f)
    elsif key == WINID_PREFIX
      win = win_names[f.getc]
      key = dec_ll(f)
    end

    t = Time.at(msec / 1000).strftime('%Y-%m-%d %X')

    modval = key >> 8
    mod = []
    MODS.each_with_index do |m, i|
      if (modval >> i) & 1 > 0
        mod << m
      end
    end

    key &= 255
    sym = KEYSYMS[key] || '???'
    puts("#{t}.%03d #{win} #{sym} #{mod*'|'}" % (msec%1000))
  end
end

